//SDL INPUT

//#include "common.h"
#include "input.h"

MouseStatus MOUSE = { 0 } ;
int INKEY = 0, SHIFT = 0, CTRL = 0, ALT = 0, S1 = 0, S2 = 0;
Controller cont[4];

#ifndef min
	#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
	#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif
#ifndef clamp
	#define clamp(small,num,big) (max((small),min((num),(big))))
#endif

#ifdef USE_SDL
#include <SDL/SDL.h>

void GetMouseEvent(SDL_Event event, MouseStatus * mou)
{
    if (event.type == SDL_MOUSEMOTION) {
        mou->x = event.motion.x;
        mou->y = event.motion.y;
    } else if (event.type == SDL_MOUSEBUTTONDOWN) {
		if (event.button.button == SDL_BUTTON_LEFT) {
			mou->b.l = true;
			mou->c.l = true;
		} else if (event.button.button == SDL_BUTTON_MIDDLE) {
			mou->b.m = true;
			mou->c.m = true;
		} else if (event.button.button == SDL_BUTTON_RIGHT) {
			mou->b.r = mou->c.r = true;
		} else if (event.button.button == 4) {
			mou->b.wu = mou->c.wu = true;
		} else if (event.button.button == 5) {
			mou->b.wd = true ; mou->c.wd = true;
		}
	} else if (event.type == SDL_MOUSEBUTTONUP) {
		if (event.button.button == SDL_BUTTON_LEFT)
			mou->b.l = false;
		else if (event.button.button == SDL_BUTTON_MIDDLE) 
			mou->b.m = false;
		else if (event.button.button == SDL_BUTTON_RIGHT) 
			mou->b.r = false;
		else if (event.button.button == 4)
			mou->b.wu = false;
		else if (event.button.button == 5)
			mou->b.wd = false;
	} 
}

void GetKeyboardEvent()
{
	return;
}
#else
#include <kos.h>

void GetMouseEvent()
{
	//ClearMouseClick(&MOUSE.c);
	memcpy(&MOUSE.last,&MOUSE.b,sizeof(MouseButtons));
	
	MAPLE_FOREACH_BEGIN(MAPLE_FUNC_MOUSE, mouse_state_t, st)
		//printf("(%i,%i) ",st->matrix[2],st->shift_keys); fflush(stdout);
		MOUSE.x = clamp(0,MOUSE.x+st->dx,640);
		MOUSE.y = clamp(0,MOUSE.y-st->dy,480);
		MOUSE.b.l = st->buttons & MOUSE_LEFTBUTTON;
		MOUSE.b.r = st->buttons & MOUSE_RIGHTBUTTON;
		MOUSE.b.m = st->buttons & MOUSE_SIDEBUTTON;
		MOUSE.c.l = MOUSE.b.l & ~MOUSE.last.l;
		MOUSE.c.r = MOUSE.b.r & ~MOUSE.last.r;
		MOUSE.c.m = MOUSE.b.m & ~MOUSE.last.m;
	MAPLE_FOREACH_END()
}

void GetKeyboardEvent()
{
	INKEY = kbd_get_key();
	MAPLE_FOREACH_BEGIN(MAPLE_FUNC_KEYBOARD, kbd_state_t, st)
		SHIFT = st->shift_keys & (KBD_MOD_LSHIFT | KBD_MOD_RSHIFT);
		CTRL = st->shift_keys & (KBD_MOD_LCTRL | KBD_MOD_RCTRL);
		ALT = st->cond.modifiers & (KBD_MOD_LALT | KBD_MOD_RALT);
		S1 = st->cond.modifiers & (KBD_MOD_S1);
		S2 = st->cond.modifiers & (KBD_MOD_S2);
	MAPLE_FOREACH_END()

	return;
}

void GetJoystickEvent()
{
	//int c = maple_first_controller();
	//TODO clean up, more controllers
	maple_device_t *contr = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);
	if (!contr)
		return;
		
	cont_state_t *cond = (cont_state_t *)maple_dev_status(contr);
	
	if (!cond)
		return;
	
	memcpy(&cont[0].last,&cont[0].b,sizeof(ControllerButtons));
	
	cont[0].b.u = cond->buttons & CONT_DPAD_UP;
	cont[0].b.d = cond->buttons & CONT_DPAD_DOWN;
	cont[0].b.l = cond->buttons & CONT_DPAD_LEFT;
	cont[0].b.r = cond->buttons & CONT_DPAD_RIGHT;
	cont[0].b.a = cond->buttons & CONT_A;
	cont[0].b.b = cond->buttons & CONT_B;
	cont[0].b.x = cond->buttons & CONT_X;
	cont[0].b.y = cond->buttons & CONT_Y;
	cont[0].b.s = cond->buttons & CONT_START;
	cont[0].b.sl = cond->ltrig > 0;
	cont[0].b.sr = cond->rtrig > 0;
	
	//TODO replace with block clicker
	cont[0].c.a = cont[0].b.a & ~cont[0].last.a;
	cont[0].c.b = cont[0].b.b & ~cont[0].last.b;
	cont[0].c.x = cont[0].b.x & ~cont[0].last.x;
	cont[0].c.y = cont[0].b.y & ~cont[0].last.y;
	cont[0].c.s = cont[0].b.s & ~cont[0].last.s;
	cont[0].c.u = cont[0].b.u & ~cont[0].last.u;
	cont[0].c.d = cont[0].b.d & ~cont[0].last.d;
	cont[0].c.l = cont[0].b.l & ~cont[0].last.l;
	cont[0].c.r = cont[0].b.r & ~cont[0].last.r;
	cont[0].c.sl = cont[0].b.sl & ~cont[0].last.sl;
	cont[0].c.sr = cont[0].b.sr & ~cont[0].last.sr;
	
	//TODO replace with block upper
	cont[0].u.a = ~cont[0].b.a & cont[0].last.a;
	cont[0].u.b = ~cont[0].b.b & cont[0].last.b;
	cont[0].u.x = ~cont[0].b.x & cont[0].last.x;
	cont[0].u.y = ~cont[0].b.y & cont[0].last.y;
	cont[0].u.s = ~cont[0].b.s & cont[0].last.s;
	cont[0].u.u = ~cont[0].b.u & cont[0].last.u;
	cont[0].u.d = ~cont[0].b.d & cont[0].last.d;
	cont[0].u.l = ~cont[0].b.l & cont[0].last.l;
	cont[0].u.r = ~cont[0].b.r & cont[0].last.r;
	cont[0].u.sl = ~cont[0].b.sl & cont[0].last.sl;
	cont[0].u.sr = ~cont[0].b.sr & cont[0].last.sr;
	
	//TODO make symetrical
	//if (cond->joyx == 0) cond->joyx = 1;
	//if (cond->joyy == 0) cond->joyy = 1;
	//cond->joyx--; cond->joyy--;
	if (cond->joyx > 0)
		cont[0].ax = ((float)cond->joyx / 127.0f);
	else
		cont[0].ax = ((float)cond->joyx / 128.0f);
	
	if (cond->joyy > 0)
		cont[0].ay = (-(float)cond->joyy / 127.0f) ;//+ 2.0f;
	else
		cont[0].ay = (-(float)cond->joyy / 128.0f) ;//+ 2.0f;
	cont[0].asl = (float)cond->ltrig / 16;
	cont[0].asr = (float)cond->rtrig / 16;
	
}
#endif

void ClearJoystickClick(ControllerButtons * b)
{
	memset(b,0,sizeof(ControllerButtons));
}

void ClearMouseClick(MouseButtons * b)
{
	b->l = b->m = b->r = b->wu = b->wd = 0;
}
