#include <kos.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include "libsh4.h"
#include "input.h"
#include "xmtrx.h"
#include "pvr_cxt.h"
#include "profile.h"
#include "watchdog.h"
#include "npvr_vertex.h"

#include "BIG_PVR.H"

#define DBG(a...)	do { printf(a); fflush(stdout); } while(0)

struct mixvertex {
	float x,y,z;
	float u;
	float nx,ny,nz;
	float v;
};

int render_strip_array_mix(void *dst, struct mixvertex *src, int stripcount, size_t *striplens, float lightx, float lighty,
	float lightz, float lightw, float ambient);

pvr_init_params_t pvr_params = {
		{ PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },
		2.6 * 1024 * 1024,
		0,	//dma
		0,	//fsaa
		0	//autosort disable
};

void pvr_setup() {
	pvr_params.fsaa_enabled = 1;
	pvr_init(&pvr_params);
	pvr_set_bg_color(0,0,0.21);
}

float __attribute__ ((aligned (32))) big_normals[obj_NumVertex * 3];
float __attribute__ ((aligned (32))) big_uvs[obj_NumVertex * 2];

typedef float fvec3[3];
void AddVec3(float *l, float *r, float *d) {
	d[0] = l[0] + r[0];
	d[1] = l[1] + r[1];
	d[2] = l[2] + r[2];
}
void SubVec3(float *l, float *r, float *d) {
	d[0] = l[0] - r[0];
	d[1] = l[1] - r[1];
	d[2] = l[2] - r[2];
}
void ModulateVec3(float *s, float v, float *d) {
	d[0] = s[0] * v;
	d[1] = s[1] * v;
	d[2] = s[2] * v;
}
void CrossVec3(float *l, float *r, float *d) {
	d[0] = l[1]*r[2] - l[2]*r[1];
	d[1] = l[2]*r[0] - l[0]*r[2];
	d[2] = l[0]*r[1] - l[1]*r[0];
}
float SquareLenVec3(float *s) {
	return s[0] * s[0] + s[1] * s[1] + s[2] * s[2];
}
void NormalizeVec3(float *s, float *d) {
	ModulateVec3(s, sh4Fsrra(SquareLenVec3(s)), d);
}

void CalculateNormals(float *pos, unsigned short *faces, int face_cnt, int vert_cnt, float *dstnorms) {
	fvec3 edge0len;
	fvec3 edge1len;
	fvec3 norm;
	int i;
	
	for(i = 0; i < face_cnt; i++, faces += 3) {
		int v0_ofs = 3 * faces[0];
		int v1_ofs = 3 * faces[1];
		int v2_ofs = 3 * faces[2];
		
		//printf("%i %i %i\n", v0_ofs, v1_ofs, v2_ofs);
		
		float *v0 = pos + v0_ofs;
		float *v1 = pos + v1_ofs;
		float *v2 = pos + v2_ofs;
		
		//Calculate normal scaled by triangle area
		SubVec3(v0, v1, edge0len);
		SubVec3(v0, v2, edge1len);
		CrossVec3(edge0len, edge1len, norm);
		
		//Sum all triangle normals contributing to vertex
		AddVec3(norm, dstnorms + v0_ofs, dstnorms + v0_ofs);
		AddVec3(norm, dstnorms + v1_ofs, dstnorms + v1_ofs);
		AddVec3(norm, dstnorms + v2_ofs, dstnorms + v2_ofs);
		//*/
	}
	
	//Normalize vertex normals
	for(i = 0; i < vert_cnt; i++) {
		NormalizeVec3(dstnorms + i*3, dstnorms + i*3);
	}
}

#define MAX_BIG_VERTS	10000
#define MAX_BIG_STRIPS	3000

struct mixvertex __attribute__ ((aligned (32))) BIGCAT[MAX_BIG_VERTS];
size_t BIGSTRIPLENS[MAX_BIG_STRIPS];
int curvertidx;

void AppendMixVertex(float x, float y, float z, float nx, float ny, float nz, float u, float v) {
	assert(curvertidx < MAX_BIG_VERTS);	//Is Big too big?
	
	BIGCAT[curvertidx].x = x;
	BIGCAT[curvertidx].y = y;
	BIGCAT[curvertidx].z = z;
	
	BIGCAT[curvertidx].nx = nx;
	BIGCAT[curvertidx].ny = ny;
	BIGCAT[curvertidx].nz = nz;
	
	BIGCAT[curvertidx].u = u;
	BIGCAT[curvertidx].v = v;
	
	curvertidx++;
}

void AppendMixVertexIdx(float *pos, float *norms, float *uvs, int idx) {
	float *po = pos + idx * 3;
	float *norm = norms + idx * 3;
	float *uv = uvs + idx * 2;
	
	AppendMixVertex(po[0], po[1], po[2], norm[0], norm[1], norm[2], uv[0], uv[1]);
}

void BuildModel(float *pos, float *norms, float *uvs, unsigned short *strips, unsigned short *striplens, int stripcnt) {
	int i, j, curstripidx = 0, stripsum = 0;
	for(j = 0; j < stripcnt; j++) {
		assert(curstripidx < MAX_BIG_STRIPS);
		
		int curstriplen = striplens[j] + 2;
		stripsum += curstriplen;
		BIGSTRIPLENS[j] = curstriplen;
		for(i = 0; i < curstriplen; i++) {
			AppendMixVertexIdx(pos, norms, uvs, *strips++);
		}
		curstripidx++;
	}
	assert(curstripidx == obj_NumStrips);
	DBG("obj_Strips len: %i\n", sizeof(obj_Strips) / sizeof(obj_Strips[0]));
	DBG("Vertices processed per model: %i\n", stripsum);
	DBG("Tris per strip: %f\n", ((float)obj_NumFaces) / obj_NumStrips);
	
}

struct mixvertex __attribute__ ((aligned (32))) testmdl[] = {
	{-5,5, 0, 0,
	0, 0, 1, 0},
	{-5,-5, 0, 0,
	0, 0, 1, 0},
	{5,5, 0, 0,
	0, 0, 1, 0},
	{5,-5, 0, 0,
	0, 0, 1, 0}
};
size_t testmdlstrip[] = {4};
	
xMatrix mtrxstack[64];


void * draw_line(float x0, float y0, float z0, float x1, float y1, float z1, float width, void *dst) {
	float px, py, t;
	pvr_vertex64s *spr = dst;
	
	//calcuate perpendicular vector of line
	px = -(y1 - y0);
	py = x1 - x0;
	
	//normalize and scale by half-width
	float square_length = px*px+py*py;
	if (square_length == 0) {
		//We have a point, not a line! Draw nothing.
		return spr;
	}
	
	t = sh4Fsrra(square_length) * width * 0.5;
	px = px*t;
	py = py*t;
	
	
	/*	example horizontal line
		clockwise winding
		a,b,c,d = screen vertex
		e = point0, f = point1
		a-------->--------b\
		|                 | |---> distance equal to abs(t)
		|e               fV/
		|                 |
		d--------<--------c
	*/
	
	pv_set_cmd_vertex_eos(spr);
	
	pv_set_pos0       (spr, x0+px, y0+py, z0);
	pv_set_pos1       (spr, x1+px, y1+py, z1);
	pv_set_submit_pos2(spr, x1-px, y1-py, z1);
	pv_set_pos3       (spr, x0-px, y0-py, z0);
	
	pv_set_uv16_0(spr, pv_pack_uv16_static(0, 0));
	pv_set_uv16_1(spr, pv_pack_uv16_static(1, 0));
	pv_set_uv16_2(spr, pv_pack_uv16_static(1, 1));
	
	return sqSubmit64BAdvance(spr);
}

int main() {
	wdInit();
	xmtrxSetStackPointer(mtrxstack, sizeof(mtrxstack));

	DBG("Setting up PVR... ");
	pvr_setup();
	DBG("done (%u bytes VRAM free)\n", pvr_mem_available());
	
	DBG("Calculating normals...\n");
	CalculateNormals(obj_Vertices, obj_Faces, obj_NumFaces, obj_NumVertex, big_normals);
	
	DBG("Building model...\n");
	BuildModel(obj_Vertices, big_normals, big_uvs, obj_Strips, obj_StripLength, obj_NumStrips);
	
	DBG("Entering render loop\n");
	
	pvr_context pc;
	pc_set_default_polygon(&pc);
	pc_set_color_format(&pc, PC_INTENSITY);
	pc_set_intensity_color(&pc, 1.0, 1.0, 0.2, 0.8);
	
	pvr_context line;
	pc_set_default_sprite(&line);
	
	pfMicroseconds submittime;
	float lastframetime_ms = 0;
	int models = 2; //18;
	void *sq;
	float avrg_render_time = 0;
	//return 0;
	do {
		wdPet();
		
		vid_border_color(0,255,0);
		GetKeyboardEvent();
		GetJoystickEvent();
		GetMouseEvent();
		
		if ((INKEY == 27) || (INKEY == KBD_KEY_DEL << 8 && CTRL && ALT) || cont[0].c.s)
			break;
		
		if (cont[0].c.r)
			models++;
		if (cont[0].c.l)
			models--;
		
		vid_border_color(0,0,0);
		pvr_wait_ready();
		vid_border_color(255,0,0);
		
		static float depth = -6;
		static short rot = 0;
		rot += 100;
		int i;
		
		submittime = pfStart();
		pvr_scene_begin();
			pvr_list_begin(PVR_LIST_OP_POLY);
				//Calculate light direction
				float lx = 0, ly = 1, lz = 0, lw = 0;
				xmtrxIdentity();
				xmtrxRotateIY(-rot * 4 + 0x2000);
				xmtrxRotateIX(-0x2000);
				sh4FtrvFV8(lx, ly, lz, lw);
				
				//Calculate perspective and save it
				xmtrxIdentity();
				if (pvr_params.fsaa_enabled) {
					mat_perspective(640.0, 240.0, 1.0, 0.2, 1000.0);
					xmtrxScale(2, 1, 1);
				} else {
					mat_perspective(320.0, 240.0, 1.0, 0.2, 1000.0);
				}
				//Save camera matrix
				xmtrxPush();
				
				//Prepare TA to recieve large cat
				sq = sqPrepare((void*)PVR_TA_INPUT);
				sqCopy32(sq, &pc);
					
				//Draw models
				for(i = 0; i < models; i++) {
					//Get clean camera matrix
					xmtrxPeek();
					
					int bigspercolumn = 6;
					float xdist = 2.75;
					float ydist = 2;
					
					//Center cat grid
					xmtrxTranslate((-(bigspercolumn-1) * xdist) / 2, -3, 0);
					
					//Place in cat grid
					xmtrxTranslate((i % bigspercolumn) * xdist, (i / bigspercolumn) * ydist, depth);
					
					//Stop arms from colliding
					xmtrxTranslate(0, (i % bigspercolumn) * ydist * 0.15, 0);
					
					//Turn model right side up
					xmtrxRotateIX(0x8000);
					
					//Spin cat
					xmtrxRotateIY(rot);
					
					//Transform, light, and submit cat
					render_strip_array_mix(sq, BIGCAT, obj_NumStrips, BIGSTRIPLENS, lx, ly, lz, 0, 0.3);
				};
				
				xmtrxDrop();
			pvr_list_finish();
			
			pvr_list_begin(PVR_LIST_TR_POLY);
				sq = sqPrepare((void*)PVR_TA_INPUT);
				pc_set_sprite_color_packed(&line, 0x80ffffff);
				sqCopy32(sq, &line);
				
				float tick_dist = pvr_params.fsaa_enabled ? 32 : 16;
				float tickwidth = pvr_params.fsaa_enabled ? 4 : 2;
				//Draw millisecond ticks
				for(i = 0; i < 17*3; i++) {
					draw_line(i * tick_dist, 470, 10,  i * tick_dist, 480, 10,  tickwidth, sq);
				}
				//Draw frame ticks
				for(i = 0; i < 4; i++) {
					draw_line(i * 16.666 * tick_dist, 460, 10,  i * 16.666 * tick_dist, 480, 10,  tickwidth, sq);
				}
				
				//CPU time bar
				pc_set_sprite_color_packed(&line, 0x8000ff00);
				sqCopy32(sq, &line);
				draw_line(0, 480, 9,  lastframetime_ms * tick_dist, 480, 9,  12, sq);
				
				//GPU time bar
				pc_set_sprite_color_packed(&line, 0x80ff0000);
				sqCopy32(sq, &line);
				draw_line(0, 480-12, 9,  avrg_render_time * tick_dist, 480-12, 9,  6, sq);
			pvr_list_finish();
		pvr_scene_finish();
		
		submittime = pfEnd(submittime);
		lastframetime_ms = pfMillisecs(submittime);
		
		pvr_stats_t stats;
		pvr_get_stats(&stats);
		avrg_render_time = avrg_render_time * 0.98 + ((float)stats.rnd_last_time) * 0.02;
		
		if (cont[0].c.a) {
			int fps = stats.frame_rate > 45 ? 60 : 30;
			DBG("\n");
			DBG("FPS: %i hz\n", fps);
			DBG("PVR render time: %.2f ms\n", avrg_render_time);
			DBG("Submit time: %.2f ms\n", pfMillisecs(submittime));
			DBG("Big count: %i cat(s)\n", models);
			DBG("Average time per Big: %.2f ms\n", pfMillisecs(submittime) / (float)models);
			DBG("Poly per frame: %i\n", obj_NumFaces * models);
			DBG("Poly per sec: %i\n", obj_NumFaces * models * fps);
			DBG("Vert buff used: %i (%i)\n", stats.vtx_buffer_used, pvr_params.vertex_buf_size);
		}
	} while(1);
	
	return 0;
}
